
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>


enum {
	_xtlfreq = 20_000_000,
	_clkfreq = 20_000_000,
	DOWNLOAD_BAUD = 230_400,
	DEBUG_BAUD = DOWNLOAD_BAUD,

	SCLpin = 13,
	SDApin = 12,
};


struct __using( "jm_i2c_modified.spin2" ) ddc;

static uint8_t  edid[128];



static void  detailedtiming( FILE *fh, int idx )
{
	int  htot, vtot, dotfreq;

	htot = edid[idx+2] | ((edid[idx+4] & 0xf0) << 4);
	vtot = edid[idx+5] | ((edid[idx+7] & 0xf0) << 4);
	fprintf( fh, "   Resolution %d x %d pixels", htot, vtot );

	htot += edid[idx+3] | ((edid[idx+4] & 0x0f) << 8);
	vtot += edid[idx+6] | ((edid[idx+7] & 0x0f) << 8);
	dotfreq = (edid[idx] | (edid[idx+1] << 8)) * 10_000;
	fprintf( fh, "   Vfreq %0.1f Hz", (float)dotfreq / (float)(htot * vtot) );

	htot = edid[idx+12] | ((edid[idx+14] & 0xf0) << 4);
	vtot = edid[idx+13] | ((edid[idx+14] & 0x0f) << 8);
	fprintf( fh, "   Dimensions %d x %d mm\n", htot, vtot );
}


static void  descriptor( FILE *fh, int idx )
{
	int  ext;
	uint8_t  ch;

	switch( edid[idx+3] )
	{
	case 0xfd :
		fprintf( fh, "   Min/Max Vertical: %d/%d Hz  %d\n", edid[idx+5], edid[idx+6], edid[idx+4] );
		fprintf( fh, "   Min/Max Horizontal: %d/%d kHz  %d\n", edid[idx+7], edid[idx+8], edid[idx+4] );
		fprintf( fh, "   Max Dot Clock: %d MHz    Supported %02x", edid[idx+9]*10, edid[idx+10] );
		break;

	case 0xff :
		fprintf( fh, "   Serial#: " );
		idx += 5;
		ext = idx + 13;
		do  {
			ch = edid[idx++];
			if( ch >= ' ' )  fputc( ch, fh );
		} while( idx < ext );
		break;

	case 0xfc :
		fprintf( fh, "   Model: " );
		idx += 5;
		ext = idx + 13;
		do  {
			ch = edid[idx++];
			if( ch >= ' ' )  fputc( ch, fh );
		} while( idx < ext );
		break;

	default :
		fprintf( fh, "   Other" );
		ext = idx + 18;
		do  {
			fprintf( fh, " %02x", edid[idx++] );
		} while( idx < ext );
	}

	fprintf( fh, "\n" );
}


static int  emitblock( FILE *fh, int daddr, int bnum )
{
	uint32_t  cksum;
	int  col, row;

	memset( edid, 0, sizeof(edid) );

	if( bnum > 1 )
	{
		ddc.start();
		ddc.write( 0x60 );
		ddc.write( bnum >> 1 );
	}
	ddc.start();
	if( !ddc.write( daddr & 0xfe ) )
	{
		if( !ddc.write( (bnum & 1) << 7 ) )    // 0 or 128
		{
			ddc.start();
			if( !ddc.write( daddr | 1 ) )  ddc.rd_block( edid, 128 );
		}
	}
	ddc.stop();

//	printf( " Device %02x   Block %02x\n", daddr, bnum );
	fprintf( fh, " Device %02x   Block %02x", daddr, bnum );
	for( row = 0; row <= 7; row++ )
	{
		fprintf( fh, "\n" );
		for( col = 0; col <= 15; col++ )
		{
			fprintf( fh, " %02x", edid[col + row*16] );
		}
	}

	cksum = 0;
	for( col = 0; col <= 127; col++ )
	{
		cksum += edid[col];
	}
	fprintf( fh, "    Checksum %02x", cksum & 0xff );
	if( cksum & 0xff || cksum == 0 )
	{
		fprintf( fh, " failed\n" );
		cksum = 1;
	} else  fprintf( fh, " passed\n" );

	return cksum;
}


static void  emit_report( FILE *fh, int daddr )
{
	int  ext, idx = 0, others = 0;

	fprintf( fh, "\n ACKs:" );
	for( idx = 0; idx < 0xff; idx += 2 )
	{
		_waitms( 1 );
		if( ddc.present( idx ) )
		{
			if( idx != daddr )  others = 1;
			fprintf( fh, " %02x", idx );
		} else  fprintf( fh, " .." );
	}
	fprintf( fh, "\n" );

	if( !emitblock( fh, daddr, 0 ) )
	{
		fprintf( fh, "   EDID Version %d.%d  Year %d", edid[0x12], edid[0x13], edid[0x11]+1990 );
		fprintf( fh, "   Display Dimensions %d x %d mm\n", edid[0x15]*10, edid[0x16]*10 );

		for( idx = 0x36; idx <= 0x6c; idx += 0x12 )
		{
			if( edid[idx] || edid[idx+1] )  detailedtiming( fh, idx );
			else  descriptor( fh, idx );
		}

		if( ext = edid[0x7e] )
		{
			for( idx = 1; idx <= ext; idx++ )  emitblock( fh, daddr, idx );
		}
	}

	if( others == 1 )
	{
		fprintf( fh, "\n Blocks of non-EDID pings\n===========================\n" );
		for( idx = 0; idx < 0xff; idx += 2 )
		{
			_waitms( 1 );
			if( idx != daddr && ddc.present( idx ) )  emitblock( fh, idx, 0 );
		}
	}

	fprintf( fh, "\n" );
}


static void  ddc_disconnect( void )
{
	int  count = 0;

	_pinclear( SCLpin );
	_pinclear( SDApin );
	_wrpin( SCLpin, P_LOW_10UA | P_HIGH_150K | P_SCHMITT_A );
	_pinl( SCLpin );    // drive pin low at 10 uA, a pull-up will hold it high
	_pinnot( 56 );    // OFF, indicates sucessful completion
	do {
		_waitms( 1 );
		if( _pinr(SCLpin) )  count = 0;    // unplugged pull-down
		else  count++;
	} while( count < 500 );    // keep looping to de-bounce

	_pinclear( SCLpin );
	_pinnot( 57 );    // OFF, indicates cable unplugged
}


static int  ddc_connect( int dchk )
{
	int  count = 0;

	_pinl( SCLpin );    // drive pin low
	_waitms( 1 );    // make sure pin is at 0 volts
	_wrpin( SCLpin, P_LOW_10UA | P_HIGH_150K | P_SCHMITT_A );  // reduce drive strength to 10 uA
	do {
		_waitms( 1 );
		if( _pinr(SCLpin) )  count++;    // plugged in pull-up rising
		else  count = 0;
	} while( count < 50 );    // keep looping to de-bounce

	_pinf( SCLpin );    // important that DIR goes low first, to avoid any small glitch
	_wrpin( SCLpin, 0 );
	_pinnot( 56 );    // ON, indicate display cable is connected

	ddc.setup( SCLpin, SDApin, 50, P_HIGH_1K5 | P_SCHMITT_A );   // 50k baud, 1.5kR pullup

	if( count = ddc.present( dchk ) )    // true if device on i2c/ddc bus
	{
		_pinnot( 57 );    // ON, indicate good-to-go
	} else {
		_pinclear( SCLpin );
		_pinclear( SDApin );
		_pinnot( 56 );    // OFF, indicate fail
	}

	return count;    // true if device on bus
}

//===================================================================================

void  main( void )
{
	char  datestr[40] = "\n ----- Wed 23 Mar 2022 -----\n";    // dafault if date file empty
	FILE  *fh;

	_pinh( 56 );    // LED indicators, OFF, on the Eval Board
	_pinh( 57 );    // LED indicators, OFF, on the Eval Board

	_clkset( 0, 25_000_000 );    // set RCFAST for low-power operation
	printf( " clkfreq = %d   clkmode = %x\n", _clockfreq(), _clockmode() );

	printf( " Mounting: " );
	mount( "/sd", _vfs_open_sdcard() );

	if( fh = fopen( "/sd/date.txt", "r" ) )
	{
		fread( datestr, 1, sizeof(datestr), fh );
		fclose( fh );
		datestr[39] = 0;
	}

	for(;;)    // loop forever
	{
		do {
			_waitms( 10 );
		} while( !ddc_connect( 0xa0 ) );    // wait until a display is plugged in

		fh = fopen( "/sd/report.txt", "a" );
		printf( " FILE = 0x%x   errno = %d\n", fh, errno );
		if( fh )
		{
			fprintf( fh, datestr );
			emit_report( fh, 0xa0 );
			fclose( fh );
			printf( " Complete!\n" );
		} else
		{
			printf( datestr );
			emit_report( stdout, 0xa0 );
			break;
		}

		ddc_disconnect();    // wait for cable removal
	}

	printf( " SD card error!\n" );
	ddc_disconnect();
	_pinh( 56 );    // OFF
	_pinh( 57 );    // OFF
}

